/*
 *
 * Author Jean-Pierre PARISY <jpparisy@free.fr>
 *
 * Copyright (C) 2013
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include "675_tuto6.h"

#define NEUTRAL 1500
#define TEMPO 15
#define PULSE 1600

static volatile uint8_t getGP0, getGP3;
static uint16_t chval[5] = {NEUTRAL,NEUTRAL,NEUTRAL,NEUTRAL,NEUTRAL};
static uint8_t outch;
static volatile bit clock;
static volatile uint16_t valGP0, valGP3, iniGP0, iniGP3;

static void interrupt routine (void);

__CONFIG(FOSC_INTRCIO & WDTE_OFF & BOREN_OFF & CP_OFF & CPD_OFF & MCLRE_OFF);

__IDLOC7('j', 'p', 'p', 6);

/* *************************************************************************************************** */

void count_sequence (uint8_t start, uint8_t stop)
{
	static uint8_t seq = 0, prev = 2;
	static uint8_t tempo = TEMPO;
	uint8_t pval = valGP0 > PULSE;

	if (prev <= 1) {							// not at initial value
		if (prev ^= pval) {						// each time the pulse value is changed,
			seq += pval;						// the sequence number is incremented
			tempo = TEMPO;
		}
		else {									// when the value doesn't change, 
			if (tempo)
				tempo--;						// wait some milliseconds before accepting the command
			else {
				chval[outch] = NEUTRAL;			// set pulse of previous active channel to neutral
				if (seq>=start && seq<=stop		// sequence inside bounds
					&& pval)					// and positive value
					outch = (seq - start) + 1;	// set ports according to the sequence number
				seq = 0;
			}
		}
	}
	prev = pval;
}

/* *************************************************************************************************** */

void decode_position (void)
{
		chval[outch] = NEUTRAL;					// set pulse of previous active channel to neutral
		if (valGP0 < 1100)
			outch = 1;							// redirect to GP1
		else if (valGP0 > 1100 && valGP0 < 1350)
			outch = 3;							// redirect to GP4
		else if (valGP0 > 1650 && valGP0 < 1900)
			outch = 4;							// redirect to GP4
		else if (valGP0 > 1900)
			outch = 2;							// redirect to GP2
		else
			outch = 0;							// pulse value around 1500, all ports to neutral
}

/* *************************************************************************************************** */

void generate_frame (void)
{
	uint8_t i;

	if (outch)									// if not neutral
		chval[outch] = valGP3;					// set pulse to active output channel
	for (i = 0; i < 4; i++) {					// write pulses on successive ports
/*										GP1      GP2     GP4     GP5		*/
		const unsigned char kport[]={0b000010,0b00100,0b010000,0b100000};
		TMR1ON = 0;								// stop timer
		clock = 0;								// reset time indicator
		TMR1 = 80 - chval[i + 1];				// set timer 1
		GPIO = kport[i];						// set port
		TMR1ON = 1;								// start timer
		while (!clock);							// wait until time has elapsed
		GPIO = 0;								// clear all ports
	}
   	getGP3 = getGP0 = 2;						// wait next pulse
   	TMR1 = 0;
}

/* *************************************************************************************************** */

void init (uint8_t val)
{
	OPTION_REG = 0b10011000;					// set OPTION register
	INTCON = 0;									// disable all interrupts and clear all flags
	T1CON = 0;									// TMR1 control register
	ADCON0 = ANSEL = 0;							// turn off analog conversion
	CMCON = 0x07; 								// turn off analog comparator
	TRISIO = 0b001001;							// GP0 and GP3 as input, all other as outputs
	GPIO = 0;									// all output ports off
	GPIF = 0;									// clear interrupt flag
	GPIE = 1;									// Port Change Interrupt enabled
	IOC = 0b001001;								// enable interrupts on GP0 or GP3 pin change
	TMR1 = 0;									// reset timer 1
	TMR1ON = 1;									// TMR1 is now counting each usec elapsed.
	TMR1IF = 0;									// clear interrupt flag
    TMR1IE = 1;									// enable timer 1 interrupt
   	PEIE = 1;									// enable peripheral interrupts
	getGP0 = getGP3 = 2;						// pulse indicator
	valGP0 = valGP3 = NEUTRAL;					// initial value
	outch = val;								// initial output channel
	ei();										// enable general interrupts
}

/* *************************************************************************************************** */

uint8_t pulse_received (void)
{
	return !getGP0 && !getGP3;
}

/* *************************************************************************************************** */

static void interrupt routine (void)
{
	if (GPIF && GPIE) {							// GPIO port change
		GPIF = 0;								// clear interrupt flag
		TMR1ON = 0;								// stop timer
		if (!GP3 && getGP3 == 1) {				// end of pulse
			valGP3 = (TMR1 - iniGP3) + 90;		// get TMR1 value
			getGP3 = 0;							// count pulse disabled
		}
		if (!GP0 && getGP0 == 1) {				// end of pulse
			valGP0 = (TMR1 - iniGP0) + 90;		// get TMR1 value
			getGP0 = 0;							// count pulse disabled
		}
		if (GP3 && getGP3 == 2) {				// start of pulse
			iniGP3 = TMR1;						// reset TMR1
			getGP3 = 1;							// count pulse in progress
		}
		if (GP0 && getGP0 == 2) {				// start of pulse
			iniGP0 = TMR1;						// reset TMR1
			getGP0 = 1;							// count pulse in progress
		}
		TMR1ON = 1;								// restart timer
	}
	if (TMR1IE && TMR1IF) {						// timer 1 interrupt
		TMR1IF = 0;								// clear interrupt flag
		clock = 1;								// set time elapsed indicator
	}
}

/* *************************************************************************************************** */
